==> oxim-1.1.5.tar.gz: oxim-1.1.5/src/xim/gui_tray.c <==
/*
    Copyright (C) 2006 by OXIM TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/Xos.h>
#include <X11/extensions/shape.h>
#include "oximtool.h"
#include "gui.h"
#include "oxim.h"

#ifdef ENABLE_EEEPC
#include "images/oxim_kbd.xpm"
#else
#include "images/oxim_tray.xpm"
#endif
int check_tray_manager(void);

static int  tray_init = False;
static char **icon = NULL;
static Atom selection_atom = None;
static Atom manager_atom = None;
static Atom system_tray_opcode_atom = None;
static Atom orientation_atom = None;
static Atom xembed_atom = None;
static Window manager_window = None;
static Window parent_window = None;
static unsigned int stamp = 0;

#define SYSTEM_TRAY_REQUEST_DOCK    0
#define SYSTEM_TRAY_BEGIN_MESSAGE   1
#define SYSTEM_TRAY_CANCEL_MESSAGE  2

static void send_message(
     Display* dpy, /* display */
     Window win,   /* sender (tray icon window) */
     long message, /* message opcode */
     long data1,   /* message data 1 */
     long data2,   /* message data 2 */
     long data3    /* message data 3 */
)
{
    XEvent ev;
  
    bzero(&ev, sizeof(ev));
    ev.xclient.type = ClientMessage;
    ev.xclient.window = win;
    ev.xclient.message_type = system_tray_opcode_atom;
    ev.xclient.format = 32;
    ev.xclient.data.l[0] = CurrentTime;
    ev.xclient.data.l[1] = message;
    ev.xclient.data.l[2] = data1;
    ev.xclient.data.l[3] = data2;
    ev.xclient.data.l[4] = data3;

//    trap_errors();
    XSendEvent(dpy, manager_window, False, NoEventMask, &ev);
    XSync(dpy, False);
/*
    if (untrap_errors()) {
    }
*/
}

/* 傳送訊息 */
static void gui_send_message(winlist_t *win)
{
    unsigned int timeout = 30000; /* 3 秒 */
    char *msg = "OXIM 輸入法"; /* message */
    int len = strlen(msg);

    stamp++;

    send_message(gui->display, /* display */
     		 manager_window,   /* sender (tray icon window) */
		 SYSTEM_TRAY_BEGIN_MESSAGE, /* message opcode */
		 timeout,   /* message data 1 */
		 len,   /* message data 2 */
		 stamp /* message data 3 */
		);

    while (len > 0)
    {
	XClientMessageEvent ev;

	ev.type = ClientMessage;
	ev.window = manager_window;
	ev.format = 8;
	ev.message_type = XInternAtom(gui->display,
			"_NET_SYSTEM_TRAY_MESSAGE_DATA", False);
	if (len > 20)
	{
	    memcpy(&ev.data, msg, 20);
	    len -= 20;
	    msg += 20;
	}
	else
	{
	    memcpy(&ev.data, msg, len);
	    len = 0;
	}
	XSendEvent(gui->display, manager_window, False, StructureNotifyMask, (XEvent *)&ev);
	XSync(gui->display, False);
    }
}

static void gui_focus_event(winlist_t *win, XEvent *event)
{
    switch (event->type)
    {
	case MapNotify:
	    DebugLog(1, ("MapNotify\n"));
	    if (!tray_init)
	    {
		XUnmapWindow(gui->display, win->window);
		alarm(1);
	    }
	    break;
	case UnmapNotify:
	    DebugLog(1, ("UnMapNotify\n"));
	    break;
	case ReparentNotify:
	    DebugLog(1, ("ReparentNotify Root(%d), Event_win(%d), Parent_Win(%d)\n", gui->root, event->xreparent.event, event->xreparent.parent));
	    parent_window = event->xreparent.parent;
	    if (parent_window == gui->root)
	    {
		DebugLog(1, ("脫離了 Manager Window\n"));
		tray_init = False;
	    }
	    break;
	case ConfigureNotify:
	    {
	    XConfigureEvent *e = &event->xconfigure;
	    DebugLog(1, ("(my %d) ConfigureNotify manager=%d, Window=%d, x=%d, y=%d, w=%d, h=%d, border_w:%d, aboveWin=%d, override=%d\n",win->window, manager_window, e->window, e->x, e->y, e->width, e->height, e->border_width, e->above, e->override_redirect));
	    int sw = (e->width - win->width) / 2;
	    int sh = (e->height - win->height) / 2;
	    int new_x = (sw > 1) ? sw : 0;
	    int new_y = (sh > 1) ? sh : 0;
	    XMoveWindow(gui->display, win->window, new_x, new_y);
	    }
	    break;
	case MotionNotify: /* Mouse Over */
	    break;
	case EnterNotify: /* Mouse In */
	    DebugLog(1, ("EnterNotify\n"));
	    //gui_send_message(win);
	    break;
	case LeaveNotify: /* Mouse Out */
	    break;
	case Expose:
	    DebugLog(1, ("Expose count=%d\n", event->xexpose.count));
	    break;
	case ButtonPress: /* 按下滑鼠按鍵 */
	    if (event->xbutton.button == Button1)
	    {
		if (gui_menu_actived())
		{
		    gui_hide_menu();
		}
		else
		{
		    gui_show_menu(win);
		}
	    }
	    break;
	case ClientMessage: /* Client message */
	    if (event->xclient.message_type == xembed_atom)
	    {
		DebugLog(1, ("in tray %d\n", event->xclient.data.l[1]));
	    }
	    else
	    {
		DebugLog(1, ("Unknow Client message -> %d\n", event->xclient.message_type));
	    }
	    break;
	case PropertyNotify:
	    DebugLog(1, ("PropertyNotify\n"));
	    break;
	case DestroyNotify:
	    DebugLog(1, ("DestroyNotify\n"));
	    break;
	default:
	    DebugLog(1, ("tray win event %d\n", event->type));
    }
}

/*
    建立 Tray 
*/
void gui_tray_init(void)
{
    winlist_t *win = NULL;

    if (gui->tray_win)
    {
	win = gui->tray_win;
    }
    else
    {
	win = gui_new_win();
    }

    /* Tray 不需要邊框 */
    XSetWindowBorderWidth(gui->display, win->window, 0);

    XSetWindowAttributes win_attr;
    win_attr.override_redirect = True;
    XChangeWindowAttributes(gui->display, win->window,
			CWOverrideRedirect, &win_attr);

    Pixmap tray, mask;
    XpmAttributes attributes;
    attributes.valuemask = 0;

    int status = XpmCreatePixmapFromData(gui->display, win->window,
			oxim_tray, &tray, &mask, &attributes);
    if (status == 0)
    {
	/* 設定長寬 */
	win->width  = attributes.width;
	win->height = attributes.height;
	/* 重設視窗大小 */
	XResizeWindow(gui->display, win->window, win->width, win->height);
	/* 設定視窗背景 */
        XSetWindowBackgroundPixmap(gui->display, win->window, tray);
	if (mask)
	{
	    XShapeCombineMask(gui->display, win->window,
			ShapeBounding, 0, 0, mask, ShapeSet);
	    XFreePixmap(gui->display, mask);
	}
	XpmFreeAttributes(&attributes);
	XFreePixmap(gui->display, tray);
    }
    win->pos_x = 0;
    win->pos_y = 0;

    win->win_draw_func = NULL;
    win->win_event_func = gui_focus_event;

    XSizeHints *size_hints = XAllocSizeHints();
    size_hints->flags = PMinSize | PMaxSize;
    size_hints->min_width = win->width;
    size_hints->min_height = win->height;
    size_hints->max_width = win->width;
    size_hints->max_height = win->height;
    XSetWMNormalHints(gui->display, win->window, size_hints);
    XFree(size_hints);

    Atom xembed_info = XInternAtom(gui->display, "_XEMBED_INFO", False);
    unsigned int mapping[2];
    mapping[0] = 0; /* ver 0 */
    mapping[1] = 1; /* request mapping */
    XChangeProperty(gui->display, win->window, xembed_info, xembed_info,
			32, PropModeReplace, (unsigned char *)mapping, 2);
    gui->tray_win = win;
}

int check_tray_manager(void)
{
    if (tray_init)
	return True;

    char tray_mgr[32];
    sprintf(tray_mgr, "_NET_SYSTEM_TRAY_S%d", gui->screen);
    selection_atom = XInternAtom(gui->display, tray_mgr, False);
    if (selection_atom == None)
    {
	return False;
    }

    manager_atom = XInternAtom(gui->display, "MANAGER", False);
    xembed_atom = XInternAtom(gui->display, "_XEMBED", False);
    system_tray_opcode_atom = XInternAtom(gui->display, "_NET_SYSTEM_TRAY_OPCODE", False);
    orientation_atom = XInternAtom(gui->display, "_NET_SYSTEM_TRAY_ORIENTATION", False);

    XGrabServer(gui->display);
    manager_window = XGetSelectionOwner(gui->display, selection_atom);

    if (manager_window != None)
    {
	XSelectInput(gui->display, manager_window, StructureNotifyMask|PropertyChangeMask);
	send_message(gui->display, manager_window, SYSTEM_TRAY_REQUEST_DOCK, gui->tray_win->window, 0, 0);
	tray_init = True;
    }
    XUngrabServer(gui->display);
    XFlush(gui->display);

    return tray_init;
}

==> oxim-1.1.6.tar.gz: oxim-1.1.6/src/xim/gui_tray.c <==
/*
    Copyright (C) 2006 by OXIM TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/Xos.h>
#include <X11/extensions/shape.h>
#include "oximtool.h"
#include "gui.h"
#include "oxim.h"

#ifdef ENABLE_EEEPC
#include "images/oxim_kbd.xpm"
#else
#include "images/oxim_tray.xpm"
#endif
static int tray_w, tray_h;

int check_tray_manager(void);

static int  tray_init = False;
static char **icon = NULL;
static Atom selection_atom = None;
static Atom manager_atom = None;
static Atom system_tray_opcode_atom = None;
static Atom orientation_atom = None;
static Atom xembed_atom = None;
static Window manager_window = None;
static Window parent_window = None;
static unsigned int stamp = 0;

#define SYSTEM_TRAY_REQUEST_DOCK    0
#define SYSTEM_TRAY_BEGIN_MESSAGE   1
#define SYSTEM_TRAY_CANCEL_MESSAGE  2

static void send_message(
     Display* dpy, /* display */
     Window win,   /* sender (tray icon window) */
     long message, /* message opcode */
     long data1,   /* message data 1 */
     long data2,   /* message data 2 */
     long data3    /* message data 3 */
)
{
    XEvent ev;
  
    bzero(&ev, sizeof(ev));
    ev.xclient.type = ClientMessage;
    ev.xclient.window = win;
    ev.xclient.message_type = system_tray_opcode_atom;
    ev.xclient.format = 32;
    ev.xclient.data.l[0] = CurrentTime;
    ev.xclient.data.l[1] = message;
    ev.xclient.data.l[2] = data1;
    ev.xclient.data.l[3] = data2;
    ev.xclient.data.l[4] = data3;

//    trap_errors();
    XSendEvent(dpy, manager_window, False, NoEventMask, &ev);
    XSync(dpy, False);
/*
    if (untrap_errors()) {
    }
*/
}

/* 傳送訊息 */
static void gui_send_message(winlist_t *win)
{
    unsigned int timeout = 30000; /* 3 秒 */
    char *msg = "OXIM 輸入法"; /* message */
    int len = strlen(msg);

    stamp++;

    send_message(gui->display, /* display */
     		 manager_window,   /* sender (tray icon window) */
		 SYSTEM_TRAY_BEGIN_MESSAGE, /* message opcode */
		 timeout,   /* message data 1 */
		 len,   /* message data 2 */
		 stamp /* message data 3 */
		);

    while (len > 0)
    {
	XClientMessageEvent ev;

	ev.type = ClientMessage;
	ev.window = manager_window;
	ev.format = 8;
	ev.message_type = XInternAtom(gui->display,
			"_NET_SYSTEM_TRAY_MESSAGE_DATA", False);
	if (len > 20)
	{
	    memcpy(&ev.data, msg, 20);
	    len -= 20;
	    msg += 20;
	}
	else
	{
	    memcpy(&ev.data, msg, len);
	    len = 0;
	}
	XSendEvent(gui->display, manager_window, False, StructureNotifyMask, (XEvent *)&ev);
	XSync(gui->display, False);
    }
}

static void gui_tray_draw(winlist_t *win)
{
    if (win->winmode & WMODE_EXIT || !tray_init)
        return;

    XpmAttributes attributes;
    Pixmap tray, mask, shapemask;
    GC shapegc, traygc;
    attributes.valuemask = 0;
    XpmCreatePixmapFromData(gui->display, win->window,
			oxim_tray, &tray, &mask, &attributes);

    tray_w = attributes.width;
    tray_h = attributes.height;

    shapemask = XCreatePixmap(gui->display, win->window,
			win->width , win->height , 1);

    traygc = XCreateGC(gui->display, tray, 0, 0);
    shapegc = XCreateGC(gui->display, shapemask, 0, 0);

    XFillRectangle(gui->display, shapemask, shapegc, 0, 0,
		win->width, win->height);

    int sw = (win->width - tray_w) / 2;
    int sh = (win->height - tray_h) / 2;
    int new_x = (sw > 0) ? sw : 0;
    int new_y = (sh > 0) ? sh : 0;

    if (mask)
    {
	XCopyArea(gui->display, mask, shapemask, shapegc, 0, 0,
		tray_w, tray_h, new_x, new_y);
	XFreePixmap(gui->display, mask);
    }
    else
    {
	XSetForeground(gui->display, shapegc, 1);
	XFillRectangle(gui->display, shapemask, shapegc, new_x, new_y,
		tray_w, tray_h);
    }

    XShapeCombineMask(gui->display, win->window, ShapeBounding,
		0, 0, shapemask, ShapeSet);
    XShapeCombineMask(gui->display, win->window, ShapeClip,
		0, 0, shapemask, ShapeSet);
    XCopyArea(gui->display, tray, win->window, traygc, 0, 0,
		win->width, win->height, new_x, new_y);

    XpmFreeAttributes(&attributes);
    XFreePixmap(gui->display, tray);
    XFreePixmap(gui->display, shapemask);
    XFreeGC(gui->display, shapegc);
    XFreeGC(gui->display, traygc);

}

static void gui_focus_event(winlist_t *win, XEvent *event)
{
    switch (event->type)
    {
	case MapNotify:
	    DebugLog(1, ("MapNotify\n"));
	    if (!tray_init)
	    {
		XUnmapWindow(gui->display, win->window);
		alarm(1);
	    }
	    break;
	case UnmapNotify:
	    DebugLog(1, ("UnMapNotify\n"));
	    break;
	case ReparentNotify:
	    DebugLog(1, ("ReparentNotify Root(%d), Event_win(%d), Parent_Win(%d)\n", gui->root, event->xreparent.event, event->xreparent.parent));
	    parent_window = event->xreparent.parent;
	    if (parent_window == gui->root)
	    {
		DebugLog(1, ("脫離了 Manager Window\n"));
		tray_init = False;
	    }
	    break;
	case ConfigureNotify:
	    {
	    XConfigureEvent *e = &event->xconfigure;
	    DebugLog(1, ("(my %d) ConfigureNotify manager=%d, Window=%d, x=%d, y=%d, w=%d, h=%d, border_w:%d, aboveWin=%d, override=%d\n",win->window, manager_window, e->window, e->x, e->y, e->width, e->height, e->border_width, e->above, e->override_redirect));
 		// width, height 會被改變，所以這裡要偵測大小是否被改變了
		// 有的話要重新把圖示放在新的範圍中間
		if (win->window == e->window && (win->width != e->width || win->height != e->height))
		{
		    DebugLog(1, ("Window size changed!\n"));
/*
		    int sw = (e->width - win->width) / 2;
		    int sh = (e->height - win->height) / 2;
		    int new_x = (sw > 1) ? sw : 0;
		    int new_y = (sh > 1) ? sh : 0;
		    XMoveWindow(gui->display, win->window, new_x, new_y);
*/
		    win->width = e->width;
		    win->height = e->height;
		}
	    }
	    break;
	case MotionNotify: /* Mouse Over */
	    break;
	case EnterNotify: /* Mouse In */
	    DebugLog(1, ("EnterNotify\n"));
	    //gui_send_message(win);
	    break;
	case LeaveNotify: /* Mouse Out */
	    break;
	case Expose:
	    DebugLog(1, ("Expose count=%d\n", event->xexpose.count));
	    if (event->xexpose.count == 0)
	    {
		gui_tray_draw(win);
	    }
	    break;
	case ButtonPress: /* 按下滑鼠按鍵 */
	    if (event->xbutton.button == Button1)
	    {
		if (gui_menu_actived())
		{
		    gui_hide_menu();
		}
		else
		{
		    gui_show_menu(win);
		}
	    }
	    break;
	case ClientMessage: /* Client message */
	    if (event->xclient.message_type == xembed_atom)
	    {
		DebugLog(1, ("in tray %d\n", event->xclient.data.l[1]));
	    }
	    else
	    {
		DebugLog(1, ("Unknow Client message -> %d\n", event->xclient.message_type));
	    }
	    break;
	case PropertyNotify:
	    DebugLog(1, ("PropertyNotify\n"));
	    break;
	case DestroyNotify:
	    DebugLog(1, ("DestroyNotify\n"));
	    break;
	default:
	    DebugLog(1, ("tray win event %d\n", event->type));
    }
}

/*
    建立 Tray 
*/
void gui_tray_init(void)
{
    winlist_t *win = NULL;

    if (gui->tray_win)
    {
	win = gui->tray_win;
    }
    else
    {
	win = gui_new_win();
    }

    /* Tray 不需要邊框 */
    XSetWindowBorderWidth(gui->display, win->window, 0);

    XSetWindowAttributes win_attr;
    win_attr.override_redirect = True;
    XChangeWindowAttributes(gui->display, win->window,
			CWOverrideRedirect, &win_attr);

    Pixmap tray, mask;
    XpmAttributes attributes;
    attributes.valuemask = 0;

    XpmCreatePixmapFromData(gui->display, win->window,
		oxim_tray, &tray, &mask, &attributes);
    tray_w = attributes.width;
    tray_h = attributes.height;
    win->width  = tray_w;
    win->height = tray_h;
    XpmFreeAttributes(&attributes);
    XFreePixmap(gui->display, tray);
    if (mask)
    {
	XFreePixmap(gui->display, mask);
    }
    win->pos_x = 0;
    win->pos_y = 0;

    win->win_draw_func = NULL;
    win->win_event_func = gui_focus_event;

    XSizeHints *size_hints = XAllocSizeHints();
    size_hints->flags = PMinSize | PMaxSize;
    size_hints->min_width = win->width;
    size_hints->min_height = win->height;
    size_hints->max_width = win->width;
    size_hints->max_height = win->height;
    XSetWMNormalHints(gui->display, win->window, size_hints);
    XFree(size_hints);

    Atom xembed_info = XInternAtom(gui->display, "_XEMBED_INFO", False);
    unsigned int mapping[2];
    mapping[0] = 0; /* ver 0 */
    mapping[1] = 1; /* request mapping */
    XChangeProperty(gui->display, win->window, xembed_info, xembed_info,
			32, PropModeReplace, (unsigned char *)mapping, 2);
    gui->tray_win = win;
}

int check_tray_manager(void)
{
    if (tray_init)
	return True;

    char tray_mgr[32];
    sprintf(tray_mgr, "_NET_SYSTEM_TRAY_S%d", gui->screen);
    selection_atom = XInternAtom(gui->display, tray_mgr, False);
    if (selection_atom == None)
    {
	return False;
    }

    manager_atom = XInternAtom(gui->display, "MANAGER", False);
    xembed_atom = XInternAtom(gui->display, "_XEMBED", False);
    system_tray_opcode_atom = XInternAtom(gui->display, "_NET_SYSTEM_TRAY_OPCODE", False);
    orientation_atom = XInternAtom(gui->display, "_NET_SYSTEM_TRAY_ORIENTATION", False);

    XGrabServer(gui->display);
    manager_window = XGetSelectionOwner(gui->display, selection_atom);

    if (manager_window != None)
    {
	XSelectInput(gui->display, manager_window, StructureNotifyMask|PropertyChangeMask);
	send_message(gui->display, manager_window, SYSTEM_TRAY_REQUEST_DOCK, gui->tray_win->window, 0, 0);
	tray_init = True;
    }
    XUngrabServer(gui->display);
    XFlush(gui->display);

    return tray_init;
}
